"""

MDiGest v.0.1.0

__version__ = 0.1.0
__author__ = Federica Maschietto <federica.maschietto@gmail.com>, Brandon Allen <bcallen95@gmail.com>

"""


from . import dim_reduction_plots
from . import general_plots
from . import plot_correlation

"""

----------

- dim_reduction plots   --> plots related to the dimensionality reduction module
- general_plots         --> plots
- plot_correlation      --> plots related to the correlation modules

"""
